/*
 * pftp -- sends files from host to host through free choosable ports
 *
 * Copyright (C) 1996-1999 Ben Schluricke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the emplied warranty of MERCHANT-
 * ABILITY OF FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *    Written by Ben Schluricke
 *    E-Mail:    bhor0533@lehr.chem.TU-Berlin.DE
 *
 * This program is dedicated to my girl-friend, Heather O'Rourke.
 *
 *
 */
#ifdef USE_POSIX_THREAD
#define _REENTRANT
#include <pthread.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>

#define SONAME    256
#define MEM_CHECK(ALLOC_MEM) \
        if (!ALLOC_MEM) { \
           fprintf(stderr, "** calloc: Cannot allocate enough memory.\n"); \
           exit(1); \
        }

const char *rcs[] = {
   "# variables",
   "PFTPPORT     1234 # enter your default port number",
   "PFTPSFILTER  gunzip:pgp -f",
   "PFTPCFILTER  gzip -4:pgp -fe",
   "PFTPPASS     * # holds the password if needed. A `*' forbids logins at all",
   "#PFTPCLIENTS  space separated list of allowed client hosts",
   "#PFTPRECEIVE  holds the absolute path to an upload directory",
   "#PFTPSLOG     holds the absolute path to a server log file",
   "#PFTPLOG      holds the absolute path to the client error log file",
   "#PFTPEDITOR   holds the name of your favorite editor",
   "#PFTPNETBUF   sets the maximum net buffer size in bytes",
   "#PFTPBANDWID  sets the bandwidth in bytes per second",
   "#PFTPSSORTBY  sets the sorting order within the pfm",
   "#PFTPUPLIMIT  sets the maximum size in kilo bytes of the upload directory\n",
   "# host names",
   "example.host.name1 2222 # port number is 2222",
   "example.host.name2      # the default port number is taken"
};

void create_rc_file()
{
   fprintf(stderr, "** Couldn't find a pftp resource file.  Create it? (y/n) ");
   if (fgetc(stdin) == 'y') {
      int i=0;
      char *rcfile=NULL;
      FILE *RCFILE;

      MEM_CHECK((rcfile = (char *)calloc(SONAME, sizeof(char))));
      if (getpwuid(getuid())) {
         sprintf(rcfile, "%s/.pftprc", (char *)getpwuid(getuid())->pw_dir);
      }
      else {
         fprintf(stderr, "Who are you?\n");
         exit(1);
      }
      if ((RCFILE = fopen(rcfile, "w+")) == NULL) {
         fprintf(stderr, "Couldn't open file `%s' for writing.\n", rcfile);
         exit(1);
      }
      for (i=0; i < 17; i++) {
         fprintf(RCFILE, "%s\n", *(rcs+i));
      }
      fclose(RCFILE);
      fprintf(stderr, "Please adjust your personal needs in `%s'.\n", rcfile);
      free(rcfile);
   }
   else {
      fprintf(stderr, "*** The resource file was not created.\n");
   }
}
