/*
 * pftp -- sends files from host to host through free choosable ports
 *
 * Copyright (C) 1996-1999 Ben Schluricke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the emplied warranty of MERCHANT-
 * ABILITY OF FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *    Written by Ben Schluricke
 *    E-Mail:    bhor0533@lehr.chem.TU-Berlin.DE
 *
 * This program is dedicated to my girl-friend, Heather O'Rourke.
 *
 *
 */
#ifdef USE_POSIX_THREAD
#define _REENTRANT
#include <pthread.h>
#endif
#if defined unicos || defined SunOS
#include <fcntl.h>
#endif
#ifdef FreeBSD
#include <sys/errno.h>
#endif
#include "main.h"

extern void set_tty(int);
extern void import(char *, unsigned long, double *, int);
extern void timing(int, double *, char *);
extern short get_var_from_pftprc(const char *, char *, short);
extern char *time_string(void);
extern short check_pass(int);
extern char *tabdir(char *);
extern void receive_udps(short strnum, double *);
#ifdef Linux
extern void handler(int);
#endif

static void test_the_net_read(FILE *, double *);
static double sum_bytes[MAXCLIENTS+1];

void receive_data(void *structure_number)
{
   char fstr[HUNAME];
   char fname[LONAME];
   char *bstrptr=NULL;
   struct stat buf;
   size_t BUFFER_SIZE = (*statstr)->_STDIN_BUFSIZ_ ? (*statstr)->_STDIN_BUFSIZ_: DEFAULT_STRING_SIZE;
#ifdef Linux
   char bstr[BUFFER_SIZE];
#else
   char bstr[BIG_BUFSIZE];
#endif
   FILE *ld;
   FILE *fp=NULL;
   char *name=NULL;
   register int cc, check;
   double filesize=0;
   double ci=0; /* number of bytes received */
   unsigned long number=0; /* This should be enough ;^) */
   int strnum = (int)*(int *)structure_number;
   int mode=0, smode=0755;
   int appending=0;
   int ccc, pp;
   int ft=1;
   unsigned char FIRST=1, hor=0, dont_write=0;
   fp = (*(statstr+strnum))->fp;

#ifdef USE_POSIX_THREAD
   pthread_mutex_lock(&interactiv);
#endif
   /*
    * If server is running as daemon or was started by inetd.
    */
   if ((*statstr)->_PFTP_DAEMON_) if (!check_pass(strnum)) {
#ifdef USE_POSIX_THREAD
      pthread_mutex_unlock(&interactiv);
      (*(statstr+strnum))->free = 1;
#endif
      return;
   }

   /*
    * Start timing.
    */
   *(sum_bytes+strnum) = 0;
   if (!(*statstr)->use_udp) {
      timing(strnum, sum_bytes+strnum, NULL);
      (*(statstr+strnum))->first = BIT_ONE;
   }

#ifdef USE_POSIX_THREAD
   pthread_mutex_unlock(&interactiv);
   if (!(*statstr)->use_udp) CHILDNUM++;
#endif

   /*
    * Handle special output.
    */
   if ((*statstr)->FORCE_STDOUT) dont_write = 1;
   if ((*statstr)->testbsize) { dont_write = 1; close(1); }
   if ((*statstr)->use_udp) dont_write = 1;

   do {
      /*
       * Get some information about the file.
       */
      *fname = '\0';
      if (!(*statstr)->use_udp) {
         if (!fgets(fstr, HUNAME, fp)) continue;
         else {
            char *strptr=NULL;
            if (((*(statstr+strnum))->first & BIT_ONE)) timing(strnum, sum_bytes+strnum, NULL);
            for (strptr=fstr; *strptr && *strptr != '\n'; strptr++);
            *strptr = '\0';
            for (cc=1; *strptr != ' ' && strptr != fstr; strptr--);
            *strptr = '\0';
            if (fstr == strptr) cc = 0;
            else {
               filesize = (double)atol(++strptr);
               if ((*(statstr+strnum))->from) {
                  if (((*(statstr+strnum))->mul -= filesize) < 0) {
                     send((*(statstr+strnum))->ws, "ERRO ** No space left.\n", 24, MSG_OOB);
                     close((*(statstr+strnum))->ws);
                     shutdown((*(statstr+strnum))->ns, 2);
                     if (slfp) {
                        fprintf(slfp, "%s %s No space left.\n",
                        time_string(), (*(statstr+strnum))->pw_name);
                        fflush(slfp);
                     }
#ifdef USE_POSIX_THREAD
                     (*(statstr+strnum))->free = 1;
#endif
                     return;
                  }
               }
               for (strptr--; *strptr != ' ' && strptr != fstr; strptr--);
               *strptr = '\0';
               if (fstr == strptr) cc = 0;
               else {
                  mode = atoi(++strptr);
                  strcpy(fname, fstr);
               }
            }

            if (!cc || ('H' != (char)fgetc(fp))) {
               fclose(fp);
               if (mode < 0) break;
               if (slfp) {
#ifdef USE_POSIX_THREAD
                  pthread_mutex_lock(&interactiv);
                  if (CHILDNUM > 1) {
                     if ((*statstr)->RECEIVING) putc('\n', slfp);
                  }
#else
                  putc('\n', slfp);
#endif
                  fprintf(slfp, "** Can't get file name from client %s\n",\
                  (*(statstr+strnum))->REMOTEHOSTNAME);
#ifdef USE_POSIX_THREAD
                  pthread_mutex_unlock(&interactiv);
#endif
               }
#ifdef USE_POSIX_THREAD
               CHILDNUM--;
               (*(statstr+strnum))->free = 1;
#endif
               return;
            }
            umask(~mode);
            umask(~smode);
         }
         if (!*fname) continue;

         if ((*(statstr+strnum))->_SKIP_ == 3) (*(statstr+strnum))->_SKIP_ = 0;
         else if ((*(statstr+strnum))->_SKIP_) (*(statstr+strnum))->_SKIP_ = 1;

#ifdef USE_POSIX_THREAD
         pthread_mutex_lock(&interactiv);
         if (slfp && FIRST && CHILDNUM == 1 && !((*statstr)->_PFTP_DAEMON_)) {
            putc(slfp == stderr ? '\r': '\n', slfp);
         }
         else if (slfp && FIRST) {
            if ((*statstr)->RECEIVING) putc('\n', slfp);
         }
#else
         if (slfp && FIRST) {
            if (!((*statstr)->_PFTP_DAEMON_)) putc('\n', slfp);
         }
#endif
         FIRST = 0;
         /*
          * Check for remote standard input stream.
          */
         if (!dont_write && (!strcmp(fname, "|")) && (*(statstr+strnum))->_STANDARD_INPUT_) {
            ft = 1;
            MEM_CHECK((name = (char *)calloc(7, sizeof(char))));
            strcpy(name, "Stream");
         }
         /*
          * Check for regular file.
          */
         else {
            /*
             * Check and create file name and directories.
             */
            if (!strcmp(fname, "|")) {
               number++;
               sprintf(fname, "%s.%lu", "Stream", number);
               /*
                * Take another number if file already exists.
                */
               while (!access(fname, F_OK)) {
                  number++;
                  sprintf(fname, "%s.%lu", "Stream", number);
               }
            }

            if ((*(statstr+strnum))->_RECURS_) {
               MEM_CHECK((name = (char *)calloc(LONAME, sizeof(char))));
            }
            else name = fname;
            for (cc=ccc=pp=0; cc < LONAME && *(fname+cc); cc++) {
               if ((*(statstr+strnum))->_RECURS_) {
                  /*
                   * Avoiding trojan horses.
                   */
                  *(name+ccc) = *(fname+cc);
                  if (*(name+ccc) == '.') pp++;
                  else if (*(name+ccc) != '/') pp = 0;

                  ccc++;
                  if (*(fname+cc) == '/') {
                     if (pp == 1) {
                        ccc -= 2;
                     }
                     else if (pp == 2) {
                        ccc -= 3;
                     }
                     else if (ccc == 1) {
                        ccc = 0;
                     }
                     else if (access(name, F_OK) && !dont_write) {
                        *(name+ccc-1) = '\0';
                        if (mkdir(name, smode)) {
                           if (slfp) fprintf(slfp, "** %s: %s\n", name, _PFTP_ERROR_ARRAY_);
                           shutdown((*(statstr+strnum))->ws, 2);
                           shutdown((*(statstr+strnum))->ns, 2);
#ifdef USE_POSIX_THREAD
                           (*(statstr+strnum))->free = 1;
                           pthread_mutex_unlock(&interactiv);
#endif
                           fclose(fp);
                           return;
                        }
                        *(name+ccc-1) = '/';
                     }
                     pp = 0;
                  }
               }
               else if (*(fname+cc) == '/') {
                  name = fname + cc + 1;
               }
            }

            /*
             * Testing for exitence of file on local host.
             */
            while (!access(name, F_OK) && !dont_write) {
               if ((*(statstr+strnum))->_SKIP_) {
                  (*(statstr+strnum))->_SKIP_ = 2;
                  break;
               }
               else if ((*(statstr+strnum))->OVERWRITE) {
                  appending = O_TRUNC;
                  if (stat(name, &buf) < 0) {
                     if (slfp) fprintf(slfp, "** stat: %s: %s\n", \
                     name, _PFTP_ERROR_ARRAY_);
                     shutdown((*(statstr+strnum))->ws, 2);
                     shutdown((*(statstr+strnum))->ns, 2);
#ifdef USE_POSIX_THREAD
                     (*(statstr+strnum))->free = 1;
                     pthread_mutex_unlock(&interactiv);
#endif
                     fclose(fp);
                     return;
                  }
                  else if ((buf.st_mode & 0200) != 0200) {
                     if (chmod(name, mode|0200) < 0) {
                        fprintf(stderr, "** chmod: %s: %s\n", \
                        name, _PFTP_ERROR_ARRAY_);
                        shutdown((*(statstr+strnum))->ws, 2);
                        shutdown((*(statstr+strnum))->ns, 2);
#ifdef USE_POSIX_THREAD
                        (*(statstr+strnum))->free = 1;
                        pthread_mutex_unlock(&interactiv);
#endif
                        fclose(fp);
                        return;
                     }
                  }
                  break;
               }
               fprintf(stderr, "** File %s exists! (from %s)\n** Overwrite(y|n) all(Y), append(a), skip(s) all(S)? ", name, (*(statstr+strnum))->REMOTEHOSTNAME);
#if defined Linux
               if ((*statstr)->ttys) {
#ifndef USE_POSIX_THREAD
                  if (!(*statstr)->single_connection) kill(getppid(), SIGUSR1);
                  set_tty(1);
#else
                  signal(SIGIO, SIG_IGN);
#endif
               }
#else
               set_tty(1);
#endif
               hor=fgetc(stdin);
#ifdef Linux
#ifndef USE_POSIX_THREAD
               if ((*statstr)->ttys) {
                  set_tty(2);
               }
#endif
#else
               set_tty(0);
#endif
               if (hor != '\n') fputc('\n', stderr);
               if(hor=='Y' || hor=='y' || hor=='a') {
                  if (stat(name, &buf) < 0) {
                     fprintf(stderr, "** stat: %s: %s\n", \
                     name, _PFTP_ERROR_ARRAY_);
                     exit(1);
                  }
                  else if ((buf.st_mode & 0200) != 0200) {
                     if (chmod(name, mode|0200) < 0) {
                        fprintf(stderr, "** chmod: %s: %s\n", \
                        name, _PFTP_ERROR_ARRAY_);
                        exit(1);
                     }
                  }
                  if (hor=='Y') (*(statstr+strnum))->OVERWRITE = BIT_ONE;
                  else if (hor=='a') appending = O_APPEND;
                  if (hor!='a') appending = O_TRUNC;
                  break;
               }
               else if(hor=='s') {
                  (*(statstr+strnum))->_SKIP_ = 3;
                  break;
               }
               else if(hor=='S') {
                  (*(statstr+strnum))->_SKIP_ = 2;
                  break;
               }
               else {
                  char *str=NULL;
                  char *stmp=NULL;
                  MEM_CHECK((str = (char *)calloc(SONAME, sizeof(char))));
#ifdef NEXTSTEP
                  getwd(str);
#else
                  getcwd(str, SONAME);
#endif
#ifndef Linux
               set_tty(2);
#endif
                  stmp = tabdir(str);
#ifndef Linux
               set_tty(0);
#endif
                  fprintf(stderr, "\n");
                  if (!stat(stmp, &buf) && (buf.st_mode & S_IFDIR)) {
                     strcat(stmp, "/");
                     strcat(stmp, name);
                  }
                  if (stmp) strcpy(name, stmp);
                  free(str);
               }   
#if defined Linux
               if ((*statstr)->ttys) {
#ifndef USE_POSIX_THREAD
                  if (!(*statstr)->single_connection) kill(getppid(), SIGUSR1);
#else
                  signal(SIGIO, handler);
#endif
                  hor = 0;
               }
#endif
            }
#if defined Linux
            if (hor) {
               hor = 0;
               if ((*statstr)->ttys) {
#ifndef USE_POSIX_THREAD
                  if (!(*statstr)->single_connection) kill(getppid(), SIGUSR1);
#else
                  signal(SIGIO, handler);
#endif
               }
            }
#endif
            if ((*(statstr+strnum))->_SKIP_ < 2 && !dont_write) {
               /*
                * Open file on local host for writing.
                */
               if ((ft = open(name, O_CREAT|O_WRONLY|appending, mode)) < 0) {
                  if ((*statstr)->_PFTP_DAEMON_ && slfp) {
                     fprintf(slfp, "%s %s: %s\n",
                     time_string(), name, _PFTP_ERROR_ARRAY_);
                  }
                  else if (slfp) {
                     fprintf(slfp, "** %s: %s\n", name, _PFTP_ERROR_ARRAY_);
                  }
                  close((*(statstr+strnum))->ws);
                  shutdown((*(statstr+strnum))->ns, 2);
#ifdef USE_POSIX_THREAD
                  (*(statstr+strnum))->free = 1;
                  pthread_mutex_unlock(&interactiv);
#endif
                  fclose(fp);
                  return;
               }
               appending = 0;
            }
            else if ((*statstr)->FORCE_STDOUT) ft = 1;
         }

         /*
          * Read file from client/filter and write to 
          * local file respectively standard out.
          */
         ci=0.0;
         (*(statstr+strnum))->lol=0;
         if (slfp && !((*statstr)->_PFTP_DAEMON_)) {
            import(name, filesize, &ci, strnum);
         }
      }
      else {
         MEM_CHECK((name = (char *)calloc(10, sizeof(char))));
         strcpy(name, "Datagrams");
      }
#ifdef USE_POSIX_THREAD
      pthread_mutex_unlock(&interactiv);
#endif

      if ((*(statstr+strnum))->_SKIP_ > 1) {
         register int rr = (unsigned int)BUFFER_SIZE < filesize ? BUFFER_SIZE : filesize;
         while (ci < filesize && (cc = fread(bstr, sizeof(char), rr, fp)) > 0) {
            if ((ci+=(double)cc) > (filesize - (double)rr)) rr = (int)(filesize - ci);
         }
      }
      else if (filesize) {
         register int rr = (unsigned int)((double)BUFFER_SIZE < filesize) ? BUFFER_SIZE : filesize;
         double firr = filesize - (double)rr;
         while (ci < filesize && (cc = fread(bstr, sizeof(char), rr, fp)) > 0) {
            if ((ci+=(double)cc) > firr) rr = (int)(filesize - ci);
            for (check=0, bstrptr=bstr; cc; cc-=check, bstrptr += check) {
               if ((check = write(ft, bstrptr, cc)) < 0) {
                  if (slfp) fprintf(slfp, "** %s: %s\n", name, _PFTP_ERROR_ARRAY_);
                  exit(1);
               }
            }
         }
      }
      else {
         if (!(*statstr)->use_udp) {
            if (!(*(statstr+strnum))->from) {
               if (!(*statstr)->testbsize) {
                  if ((*statstr)->_STDIN_BUFSIZ_) {
                     register int rr = BUFFER_SIZE;
                     for (ci=0, cc=0; (cc = fread(bstr, sizeof(char), rr, fp)) > 0;) {
                        ci+=(double)cc;
                        for (check=0, bstrptr=bstr; cc; cc-=check, bstrptr += check) {
                           if ((check = write(ft, bstrptr, cc)) < 0) {
                              if (slfp) fprintf(slfp, "** %s: %s\n", name, _PFTP_ERROR_ARRAY_);
                              exit(1);
                           }
                        }
                     }
                  }
                  else {
                     for (ci=0.0; (cc = fgetc(fp)) != EOF; ci++) fputc(cc, stdout);
                  }
               }
               else test_the_net_read(fp, &ci);
            }
            else {
               register int rr = BUFFER_SIZE;
               for (ci=0.0, cc=0; (cc = fread(bstr, sizeof(char), rr, fp)) > 0;) {
                  if (((*(statstr+strnum))->mul -= (double)cc) < 0) {
                     send((*(statstr+strnum))->ws, "ERRO ** No space left.\n", 24, MSG_OOB);
                     close((*(statstr+strnum))->ws);
                     shutdown((*(statstr+strnum))->ns, 2);
                     if (slfp) {
                        fprintf(slfp, "%s %s No space left.\n",
                        time_string(), (*(statstr+strnum))->pw_name);
                        fflush(slfp);
                     }
#ifdef USE_POSIX_THREAD
                     (*(statstr+strnum))->free = 1;
#endif
                     return;
                  }
                  ci+=(double)cc;
                  for (check=0, bstrptr=bstr; cc; cc-=check, bstrptr += check) {
                     if ((check = write(ft, bstrptr, cc)) < 0) {
                        if (slfp) fprintf(slfp, "** %s: %s\n", name, _PFTP_ERROR_ARRAY_);
                        exit(1);
                     }
                  }
               }
            }
         }
         else receive_udps(strnum, &ci);
      }

      if (!ci) (*(statstr+strnum))->lol = 1;
      else *(sum_bytes+strnum) += ci;
      if ((*(statstr+strnum))->_SKIP_ < 2 && !(*statstr)->FORCE_STDOUT && !(*(statstr+strnum))->_STANDARD_INPUT_) close(ft);
#ifdef USE_POSIX_THREAD
      pthread_mutex_lock(&interactiv);
#endif
      if (slfp && (!((*statstr)->_PFTP_DAEMON_) || (*statstr)->lognames)) {
         if (filesize) import(name, filesize, &ci, strnum);
         else import(name, filesize, sum_bytes+strnum, strnum);
      }

      if ((*(statstr+strnum))->_RECURS_) free(name);
      /*
       * Check of size of file.
       */
      if (slfp && filesize > ci && filesize > 0.0) {
         fprintf(slfp, "** File is incomplete. ");
         fprintf(slfp, "Actual size of file is %.0f bytes.\n", filesize);
         if (pftplog) {
            if ((ld = fopen(pftplog, "a")) != NULL) {
               fprintf(ld, "** File is incomplete. ");
               fprintf(ld, "Actual size of file is %.0f bytes.\n", filesize);
               fclose(ld);
            }
         }
      }
#ifdef USE_POSIX_THREAD
      pthread_mutex_unlock(&interactiv);
#endif
      ci = 0.0;
   } while (strcmp(fname, ""));
#ifdef USE_POSIX_THREAD
   pthread_mutex_lock(&interactiv);
   if (slfp) {
      if ((*statstr)->RECEIVING && !(*(statstr+strnum))->_STANDARD_INPUT_) {
         putc('\n', slfp);
      }
      if ((*statstr)->_PFTP_DAEMON_) {
         fprintf(slfp, "%s %s", time_string(),
         (*(statstr+strnum))->REMOTEHOSTNAME);
      }
      else {
         fprintf(slfp, "<%d> Connection to %s",
         strnum, (*(statstr+strnum))->REMOTEHOSTNAME);
      }
   }
   /*
    * End timing.
    */
   timing(strnum, sum_bytes+strnum, fstr);
   
   pthread_mutex_unlock(&interactiv);
   (*(statstr+strnum))->free = 1;
   if (CHILDNUM) CHILDNUM--;
#else
   if (slfp) {
      if ((*statstr)->_PFTP_DAEMON_) {
         fprintf(slfp, "%s %s", time_string(),
         (*(statstr+strnum))->REMOTEHOSTNAME);
      }
      else {
         fprintf(slfp, "<*> Connection to %s",
         (*(statstr+strnum))->REMOTEHOSTNAME);
      }
   }
   /*
    * End timing.
    */
   timing(strnum, sum_bytes+strnum, fstr);
#endif
   if (!(*statstr)->use_udp) {
      if (write((*(statstr+strnum))->ws, fstr, strlen(fstr)) < 0) {
         fprintf(slfp, "** write: %s\n", _PFTP_ERROR_ARRAY_);
      }
      if ((*(statstr+strnum))->pw_name) free((*(statstr+strnum))->pw_name);
      if ((*(statstr+strnum))->from) free((*(statstr+strnum))->from);
      shutdown((*(statstr+strnum))->ws, 2);
      shutdown((*(statstr+strnum))->ns, 2);
   }
}


static void test_the_net_read(FILE *fp, double *ci)
{
   size_t cc=0, bufsize = (*statstr)->_STDIN_BUFSIZ_;
   char *buf = malloc(bufsize*sizeof(char));
   
   for (*ci=cc=0; (cc = fread(buf, sizeof(char), bufsize, fp)) > 0; *ci+=(double)cc);
   free(buf);
}
