/*
 * pftp -- sends files from host to host through free choosable ports
 *
 * Copyright (C) 1996-1999 Ben Schluricke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the emplied warranty of MERCHANT-
 * ABILITY OF FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *    Written by Ben Schluricke
 *    E-Mail:    bhor0533@lehr.chem.TU-Berlin.DE
 *
 * This program is dedicated to my girl-friend, Heather O'Rourke.
 *
 *
 */
#ifdef USE_POSIX_THREAD
#define _REENTRANT
#include <pthread.h>
#endif
#include <time.h>
#if defined HAVE_GLIB || defined OSF1
#include <sys/time.h>
#endif
#include "main.h"

/*
 * Get Client's hostname.
 */
void import(char *filename, unsigned long filesize, double *ci, int strnum)
{
#ifdef USE_POSIX_THREAD
   if (!*ci) {
      OUTPUT++;
      (*statstr+strnum)->output = OUTPUT;
      (*statstr)->RECEIVING = 1;
   }
#endif

#ifdef USE_POSIX_THREAD
   if ((*(statstr+strnum))->lol) {
      if ((*statstr+strnum)->output == OUTPUT) fputc(slfp == stderr ? '\r': '\n', slfp);
      else if ((*statstr)->RECEIVING) fputc('\n', slfp);
      (*statstr)->RECEIVING = 0;
   }
#else
   if ((*(statstr+strnum))->lol) fputc('\r', slfp);
#endif
   fflush(slfp);
   if (*ci > 0.0F) fprintf(slfp, "+ %s (%.0f bytes) %s from %s.\n", filename, *ci, (*(statstr+strnum))->_SKIP_ > 1 ? "skipped": "received", (*(statstr+strnum))->REMOTEHOSTNAME);
   else if (!*ci && (*(statstr+strnum))->lol) {
      fprintf(slfp, "\n** Receiving %s from %s failed.\n", filename, (*(statstr+strnum))->REMOTEHOSTNAME);
      if (pftplog) {
         FILE *fe=slfp;
         if ((fe = fopen(pftplog, "a")) != NULL) {
            fprintf(fe, "\n** Receiving %s from %s failed.\n", filename, (*(statstr+strnum))->REMOTEHOSTNAME);
            fclose(fe);
         }
      }
   }
   else {
      if (!(*(statstr+strnum))->_STANDARD_INPUT_ || *filename != '|') (*(statstr+strnum))->lol = 1;
      if (slfp == stderr) {
         if (filesize) fprintf(slfp, "* %s %s (%ld bytes) from %s", (*(statstr+strnum))->_SKIP_ > 1 ? "Skipping": "Receiving", filename, filesize, (*(statstr+strnum))->REMOTEHOSTNAME);
         else fprintf(slfp, "* %s %s from %s", (*(statstr+strnum))->_SKIP_ > 1 ? "Skipping": "Receiving", filename, (*(statstr+strnum))->REMOTEHOSTNAME);
      }
   }
   if (!*ci && (((*(statstr+strnum))->_STANDARD_INPUT_ && !filesize) || (*statstr)->FORCE_STDOUT)) fprintf(slfp, "\n");
   fflush(slfp);
}

void timing(int strnum, double *sum_bytes, char *str)
{
#ifdef USE_POSIX_THREAD
   double start;
#else
   static double start;
#endif
   double diff=0;
   double rate=0;
#ifdef OSF1
   struct {
      time_t tv_sec;
      int tv_usec;
   } tv;
#else
   struct {
      long tv_sec;
      long tv_usec;
   } tv;
#endif

   if (!(*(statstr+strnum))->first)
   {
      gettimeofday((struct timeval *)&tv, (struct timezone *)NULL);
      start = (double) tv.tv_sec + ((double) tv.tv_usec / 1000000.0F);
#ifdef USE_POSIX_THREAD
      (*(statstr+strnum))->start = start;
#endif
   }
   else if (((*(statstr+strnum))->first & BIT_ONE)) {
      /*
       * Assuming the time it takes for the first datagram
       * sent till it is received is half the time the ACK
       * is sent back and the first datagram is sent.
       */
      (*(statstr+strnum))->first = BIT_TWO; 
#ifdef USE_POSIX_THREAD
      start = (*(statstr+strnum))->start;
#endif
      gettimeofday((struct timeval *)&tv, (struct timezone *)NULL);
      diff   = (double) tv.tv_sec + ((double) tv.tv_usec / 1000000.0F);
      diff   = (diff - start)/(double)2;
      if ((*statstr)->use_udp) start -= diff;
      else start += diff;
   }
   else {
#ifdef USE_POSIX_THREAD
      start = (*(statstr+strnum))->start;
#endif
      gettimeofday((struct timeval *)&tv, (struct timezone *)NULL);
      diff  = (double) tv.tv_sec + ((double) tv.tv_usec / 1000000.0F);
      diff -= start;
      rate  = *sum_bytes / diff;
      rate /= (double)1024;
      if (slfp) {
         if ((*statstr)->_PFTP_DAEMON_) {
            fprintf(slfp, " (%ldB %.2fs %.2fKB/s)\n", (unsigned long)*sum_bytes, diff, rate);
         }
         else {
            fprintf(slfp, " (%ldB %.2fs %.2fKB/s) closed.\n\r", (unsigned long)*sum_bytes, diff, rate);
         }
      }
      if (str) sprintf(str, "OKAY <*> Sent %ld bytes in %.2f seconds (%.2fKB/s)", (unsigned long)*sum_bytes, diff, rate);
   }
}

char *time_string(void)
{
   time_t secs=1;
   char *str=NULL, *rstr=NULL;

   time(&secs);
   str = asctime(localtime(&secs))+4;
   for (rstr=str; *rstr && *rstr != '\n'; rstr++);
   *rstr = '\0';

   return str;
}


/* This function is executed by the client */
void check_bandwidth(size_t sbytes)
{
   static double ct[2] = {0, 0};
   static double bytes[2] = {0, 0};
   static int num=0;
   double current_time=0;
   double maxsize=0;
   double ttw=0; /* time to wait */
#ifdef OSF1
   struct {
      time_t tv_sec;
      int tv_usec;
   } tv;
#else
   struct {
      long tv_sec;
      long tv_usec;
   } tv;
#endif
      if (!(*statstr)->_BANDWIDTH_) return;
      gettimeofday((struct timeval *)&tv, (struct timezone *)NULL);
      /* in microseconds */
      current_time = ((double) tv.tv_sec * 1000000.0F) + (double) tv.tv_usec;
   if (*(ct+num)) {
      *(bytes+num) += (double)sbytes;
      maxsize = (*statstr)->_BANDWIDTH_ * 2;
      /* in microseconds */
      ttw = ((*(bytes+num) * 1000000.0F) / ((double)(*statstr)->_BANDWIDTH_)) - (current_time - *(ct+num));
#ifdef HAVE_USLEEP
      if (ttw > 0) usleep((unsigned long)ttw);
#else
      if (ttw > 0) {
         tv.tv_sec = ttw / 1000000;
         tv.tv_usec = ((unsigned long)ttw) % 1000000;
         select(0, NULL, NULL, NULL, (struct timeval *)&tv);
      }
#endif
      if (*(bytes+num) > maxsize) {
         if (*(ct+(!num))) *(bytes+(!num)) += (double)sbytes;
         else {
            gettimeofday((struct timeval *)&tv, (struct timezone *)NULL);
            /* in microseconds */
            current_time = ((double) tv.tv_sec * 1000000.0F) + (double) tv.tv_usec;
            *(ct+(!num)) = current_time;
         }
      }
      if (*(bytes+num) > maxsize * 2) {
         *(bytes+num) = 0;
         *(ct+num) = 0;
         num = !num;
      }
   }
   else {
      *(ct+num) = current_time;
      *(bytes) = (double)0;
      *(bytes+1) = (double)0;
   }
}
