/*
 * pftp -- sends files from host to host through free choosable ports
 *
 * Copyright (C) 1996-1999 Ben Schluricke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the emplied warranty of MERCHANT-
 * ABILITY OF FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *    Written by Ben Schluricke
 *    E-Mail:    bhor0533@lehr.chem.TU-Berlin.DE
 *
 * This program is dedicated to my girl-friend, Heather O'Rourke.
 *
 *
 */
#ifdef USE_POSIX_THREAD
#define _REENTRANT
#include <pthread.h>
#endif
#ifdef FreeBSD
#include <sys/errno.h>
#endif
#ifdef AIX
#include <sys/select.h>
#include <values.h>
#endif
#if !defined HP_UX
#include <fcntl.h>
#endif
#include <sys/wait.h>
#include <sys/types.h>
#include <pwd.h>
#ifdef HAS_SHADOW_PASSWD
#include <shadow.h>
#endif
#include "main.h"
#ifdef unicos
#define PFTP_THREAD_CAST (void *(*)(void *))
#else
#define PFTP_THREAD_CAST (void *)
#endif

extern void accs(int, int, char *);
extern void set_tty(int);
extern void handler(int);
extern void server_filter_programm(void *);
extern void receive_data(void *);
extern void free_memory(void);
extern int init_struct(void);
extern short get_var_from_pftprc(const char *, char *, short);
extern short check_login(int, int);
#if defined Linux && !defined USE_POSIX_THREAD
extern void user_handler(int);
#endif
#if !defined USE_POSIX_THREAD
extern void clients_handler(int);
#endif
/*
 * Server--This is the server part of pftp.
 */
void Server(int portn)
{
   char hostname[SONAME];
   FILE *fp=NULL;
   int ns=0, ws=0, optval=(*statstr)->bsize;
#ifdef HAVE_INET6
   struct sockaddr_in6 fsin;
   int ret=0;
#else
   struct hostent *hp;
   struct sockaddr_in fsin;
#endif
#ifndef USE_POSIX_THREAD
#ifndef AIX
   fd_set rfds;
#else
   SELLIST(BITS(int), BITS(int)) rfds;
#endif
#ifdef OSF1
   struct {
      time_t tv_sec;
      int tv_usec;
   } tv;
#else
   struct {
      long tv_sec;
      long tv_usec;
   } tv;
#endif
#endif
   int fsinlen=sizeof(fsin);
   int pid=0;
   int strnum=0;
#if !defined USE_POSIX_THREAD
   int child_status=0;
#endif
#ifdef USE_POSIX_THREAD
   pthread_t thr;
   pthread_attr_t thrattr;

   pthread_attr_init(&thrattr);
   pthread_attr_setdetachstate(&thrattr, PTHREAD_CREATE_DETACHED);
#endif

   /*
    * Set signals.
    */
   signal(SIGQUIT, handler);
   signal(SIGINT, handler);
#if defined Linux
   if (!(*statstr)->_PFTP_DAEMON_ && (*statstr)->ttys) {
      /*
       * Special terminal handle for Linux
       * since Linux supports the FASYNC
       * for a normal tty! Great! :^)
       */
      signal(SIGIO, handler);
      set_tty(2);
      fcntl(0, F_SETOWN, getpid());
      pid = fcntl(0, F_GETFL);
      fcntl(0, F_SETFL, pid | FASYNC);
#ifndef USE_POSIX_THREAD
      signal(SIGUSR1, user_handler);
#endif
   }
#endif

   /*
    * Set stderr.
    */
   if (!slfp) {
      char *slogname = NULL;
      if ((slogname = getenv("PFTPSLOG")));
      else {
         MEM_CHECK((slogname = (char *)calloc(SONAME, sizeof(char))));
         if (!get_var_from_pftprc("PFTPSLOG", slogname, 0)) {
            free(slogname);
            slogname = NULL;
         }
      }
      if (slogname) {
         if ((slfp = fopen(slogname, "a")) == NULL) {
            fprintf(stderr, "** %s: %s\n", slogname, _PFTP_ERROR_ARRAY_);
            exit(1);
         }
         chmod(slogname, 0600);
         if (!getenv("PFTPSLOG")) free(slogname);
      }
      if (!(*statstr)->OVERWRITE) (*statstr)->_SKIP_ = 1;
   }

   /*
    * Started by inetd.
    */
   if (((*statstr)->_PFTP_DAEMON_ & BIT_TWO)) {
      (*statstr)->exit_afc = 1;
      s = dup(0);
      dup2(ns, 0);
      dup2(ws, 2);
      close(1);
   }
   else {
      /*
       * Initialize socket s, and perform some
       * socket options.  Eventually bind our
       * address to the socket.
       */
      accs((short)1, portn, hostname);
   }
   if ((*statstr)->exit_afc || (*statstr)->use_udp) {
      (*statstr)->single_connection = 1;
   }
   if (!(*statstr)->use_udp && !((*statstr)->_PFTP_DAEMON_ & BIT_TWO)) {
      /*
       * Listen on the socket.
       */
      if (listen(s, (*statstr)->single_connection ? 1: (*statstr)->maxclients) < 0) {
         if (slfp) fprintf(slfp, "** listen: %s\n", _PFTP_ERROR_ARRAY_);
         exit(1);
      }
   }

   while (s) {
      if (!(*statstr)->use_udp) {
         /* *((*statstr)->REMOTEHOSTNAME) = '\0';  */
         if ((*statstr)->OVERWRITE==1) (*statstr)->OVERWRITE=0;
         if (!((*statstr)->_PFTP_DAEMON_ & BIT_TWO)) {
#if !defined USE_POSIX_THREAD
            do {
#ifndef AIX
               FD_ZERO(&rfds);
               FD_SET(s, &rfds);
#else
               /* The follwoing is just written as I understood
                * the AIX select manual and it works.  If you
                * know more about this construct, please drop
                * me an e-mail.  Thanks in advance!
                */
               rfds.fdsmask[s / BITS(int)] = 1 << (s % BITS(s));
#endif
               tv.tv_sec = 1;
               tv.tv_usec = 0;
               /*
                * See if childs exited.
                */
               if (waitpid((pid_t)-1, &child_status, WNOHANG) > 0) (*statstr)->no_clients--;
            }
#ifndef AIX
            while (select(s+1, &rfds, NULL, NULL, (struct timeval *)&tv) <= 0);
#else
            while (select(BITS(s), &rfds, NULL, NULL, (struct timeval *)&tv) <= 0);
#endif
#endif
            /*
             * Accept connections.  When we accept one, ns
             * will be connected to the client.
             */
            if ((ns = accept(s, (struct sockaddr *)&fsin, &fsinlen)) < 0) {
               if (slfp) fprintf(slfp, "** accept: %s\n", _PFTP_ERROR_ARRAY_);
               exit(1);
            }
            ws = ns;
         }
         else {
            if (getpeername(ns, (struct sockaddr *)&fsin, (int *)&fsinlen) < 0) {
               if (slfp) fprintf(slfp, "** getsockname: %s\n", _PFTP_ERROR_ARRAY_);
               exit(1);
            }
         }

         /*
          * Get client's host name.
          */
#ifdef HAVE_INET6
         if ((ret = getnameinfo((struct sockaddr *)&fsin, fsinlen, \
            (*statstr)->REMOTEHOSTNAME, SONAME*sizeof(char), NULL, 0, 0))) {
            if (!inet_ntop(fsin.sin6_family, fsin.sin6_addr.s6_addr, (*statstr)->REMOTEHOSTNAME, INET6_ADDRSTRLEN)) {
               fprintf(slfp, "Cannot convert binary address string.\n");
               if (shutdown(ns, 2) < 0) {
                  if (slfp) fprintf(slfp, "** shutdown: %s\n", _PFTP_ERROR_ARRAY_);
                  exit(1);
               }
            }
         }
#else
         if ((hp = gethostbyaddr((char *)&fsin.sin_addr, sizeof(fsin.sin_addr), AF_INET)) == 0) {
            strcpy((*statstr)->REMOTEHOSTNAME, inet_ntoa(fsin.sin_addr));
         }
         else {
            strcpy((*statstr)->REMOTEHOSTNAME, hp->h_name);
         }
#endif

         /*
          * Check if remote server has a valid host name.
          */
         if (*_CLIENTHOSTNAME_) {
            int i, loc=0, cl, isok=0, length = strlen((*statstr)->REMOTEHOSTNAME);

            for (cl=0; _CLIENTHOSTNAME_[cl] && !isok; cl++) {
               loc = strlen(*(_CLIENTHOSTNAME_+cl));

               if (length < loc) continue;
               else if (*_CLIENTHOSTNAME_[cl] == '.')
               {
                  loc--;
                  for (i=length-1; loc; loc--, i--) {
                     if (*(_CLIENTHOSTNAME_[cl]+loc) != tolower(*((*statstr)->REMOTEHOSTNAME+i))) break;
                  }
                  if (!loc) isok = 1;
               }
               else if (length == loc) {
                  isok=1;
                  for (i=0; i < length; i++) {
                     if (*(_CLIENTHOSTNAME_[cl]+i) != tolower(*((*statstr)->REMOTEHOSTNAME+i))) {
                        isok = 0;
                        break;
                     }
                  }
               }
            }
            if (!isok) {
               if (slfp) fprintf(slfp, "** Connection to %s refused.\n", (*statstr)->REMOTEHOSTNAME);
               if (shutdown(ns, 2) < 0) {
                  if (slfp) fprintf(slfp, "** shutdown: %s\n", _PFTP_ERROR_ARRAY_);
                  exit(1);
               }
               if ((*statstr)->exit_afc) {
                  if (shutdown(s, 2) < 0) {
                     if (slfp) fprintf(slfp, "** shutdown: %s\n", _PFTP_ERROR_ARRAY_);
                     exit(1);
                  }
                  s = 0;
                  break;
               }
               continue;
            }
         }
         if (*((*statstr)->REMOTEHOSTNAME) == '.') {
            *((*statstr)->REMOTEHOSTNAME) = ' ';
            *((*statstr)->REMOTEHOSTNAME+7) = ' ';
         }

         if (setsockopt(ns, SOL_SOCKET, SO_RCVBUF, (char *)&optval, sizeof(optval)) < 0) {
            if (slfp) fprintf(slfp, "** setsockopt: %s\n", _PFTP_ERROR_ARRAY_);
            exit(1);
         }
      }
      else ns = s;
#ifdef USE_POSIX_THREAD
      strnum = init_struct();
      if (!strnum) {
         if (slfp) fprintf(slfp, "\n** Connection to %s refused.\n", (*statstr)->REMOTEHOSTNAME);
         if (shutdown(ns, 2) < 0) {
            if (slfp) fprintf(slfp, "** shutdown: %s\n", _PFTP_ERROR_ARRAY_);
            (*(statstr+strnum))->free = 1;
            exit(1);
         }
         if (((*statstr)->_PFTP_DAEMON_ & BIT_TWO)) s = 0;
         (*(statstr+strnum))->free = 1;
         continue;
      }
#else
      strnum = 0;
      if (!(*statstr)->single_connection) {
         (*statstr)->no_clients++;
         if ((*statstr)->no_clients > (*statstr)->maxclients) {
            (*statstr)->no_clients--;
            if (slfp) fprintf(slfp, "\n** Connection to %s refused.\n", (*statstr)->REMOTEHOSTNAME);
            if (shutdown(ns, 2) < 0) {
               if (slfp) fprintf(slfp, "** shutdown: %s\n", _PFTP_ERROR_ARRAY_);
               exit(1);
            }
            if (((*statstr)->_PFTP_DAEMON_ & BIT_TWO)) s = 0;
            continue;
         }
      }
      (*statstr)->first = 0;
#endif
      if (!(*statstr)->use_udp) {
         (*(statstr+strnum))->ns = dup(ns);
         (*(statstr+strnum))->ws = dup(ws);
         if (write((*(statstr+strnum))->ws, "PFTP ", 6) < 0) {
            if (slfp) fprintf(slfp, "** write: %s\n", _PFTP_ERROR_ARRAY_);
         }
         if (write((*(statstr+strnum))->ws, VERSION, strlen(VERSION)) < 0) {
            if (slfp) fprintf(slfp, "** write: %s\n", _PFTP_ERROR_ARRAY_);
         }
         if (write((*(statstr+strnum))->ws, "\n", 1) < 0) {
            if (slfp) fprintf(slfp, "** write: %s\n", _PFTP_ERROR_ARRAY_);
         }

         /*
          * If running as a daemon or
          * server started by inetd...
          */
         if ((*statstr)->_PFTP_DAEMON_) {
            if (!check_login(ns, strnum)) {
               if ((*statstr)->_PFTP_DAEMON_ & BIT_TWO) {
                  shutdown(s, 2);
                  s = 0;
               }
               continue;
            }
         }
         else {
            /*
             * Send status to the client.
             */
            if (write((*(statstr+strnum))->ws, "STAT server\n", 13) < 0) {
               if (slfp) fprintf(slfp, "** write: %s\n", _PFTP_ERROR_ARRAY_);
            }
            MEM_CHECK(((*(statstr+strnum))->home = (char *)calloc(SONAME, sizeof(char))));
            strcpy((*(statstr+strnum))->home, getenv("HOME"));
            strcat((*(statstr+strnum))->home, PFTPRESOURCE);
         }
         if (slfp) fflush(slfp);
      }

      if (*filter) {
         if (slfp && !(*statstr)->_PFTP_DAEMON_) fprintf(slfp, "\n* Filter is receiving data from %s ...\n", (*statstr)->REMOTEHOSTNAME);
         if (!(*statstr)->single_connection) {
            /*
             * Unfortunately gettimeofday is neither in
             * libc5 nor in glibc thread-safe.
             */
#ifdef USE_POSIX_THREAD
            /*
             * Create a new thread for the filter routine.
             */
            if ((pid = pthread_create(&thr, &thrattr, PFTP_THREAD_CAST &server_filter_programm, (void *)&strnum)) != 0) {
               if (slfp) fprintf(slfp, "** pthread_create: %s\n", _PFTP_ERROR_ARRAY_PID_);
               exit(1);
            }
#else
            /*
             * Spawn a new process for the filter routine.
             */
            if ((pid = fork()) < 0) {
               if (slfp) fprintf(slfp, "** fork: %s\n", _PFTP_ERROR_ARRAY_);
               exit(1);
            }

            /*
             * Here starts the filter process.
             */
            if (!pid) {
               (*statstr)->isparent = 0;
               server_filter_programm((void *)&strnum);
               s = 0;
            }
#endif /* USE_POSIX_THREAD */
         }
         /*
          * Filtered standard input/output or '-i' specified.
          */
         else {
            server_filter_programm((void *)&strnum);
         }
      }
      else if (!(*statstr)->single_connection) {
#ifdef USE_POSIX_THREAD
         fp = fdopen(ns, "r");
         (*(statstr+strnum))->fp = fp;
         if ((pid = pthread_create(&thr, &thrattr, PFTP_THREAD_CAST &receive_data, (void *)&strnum)) != 0) {
            if (slfp) fprintf(slfp, "** pthread_create: %s\n", _PFTP_ERROR_ARRAY_PID_);
            exit(1);
         }
#else
         if ((pid = fork()) < 0) {
            if (slfp) fprintf(slfp, "** fork: %s\n", _PFTP_ERROR_ARRAY_);
            exit(1);
         }
         if (!pid) {
            signal(SIGCHLD, SIG_DFL);
            signal(SIGQUIT, SIG_DFL);
            signal(SIGINT, SIG_DFL);
#if defined Linux
            signal(SIGIO, SIG_DFL);
            signal(SIGUSR1, SIG_DFL);
#endif
            fp = fdopen(ns, "r");
            (*(statstr+strnum))->fp = fp;
            (*statstr)->isparent = 0;
            receive_data((void *)&strnum);
            fclose(fp);
            s = 0;
         }
#endif
      }
     /*
      * Standard input/output or '-i' specified.
      */
      else {
         if (!(*statstr)->use_udp) {
            fp = fdopen(ns, "r");
            (*(statstr+strnum))->fp = fp;
         }
         receive_data((void *)&strnum);
      }
      if ((*statstr)->exit_afc) s = 0;
   }
   shutdown(s, 2);
   free_memory();

   exit(1);
}
