/*
 * pftp -- sends files from host to host through free choosable ports
 *
 * Copyright (C) 1996-1999 Ben Schluricke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the emplied warranty of MERCHANT-
 * ABILITY OF FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *    Written by Ben Schluricke
 *    E-Mail:    bhor0533@lehr.chem.TU-Berlin.DE
 *
 * This program is dedicated to my girl-friend, Heather O'Rourke.
 *
 *
 */
#ifndef STATUS_STRUCTURE_H
#define STATUS_STRUCTURE_H

#if !defined Linux
#include <errno.h>
#endif
#if !defined HAVE_STRERROR
#define _PFTP_ERROR_ARRAY_     "error"
#define _PFTP_ERROR_ARRAY_PID_ "error"
#else
#define _PFTP_ERROR_ARRAY_     strerror(errno)
#define _PFTP_ERROR_ARRAY_PID_ strerror(pid)
#endif

#include "connect.h"

#define BIT_ONE              01
#define BIT_TWO              02
#define PFTPRESOURCE         "/.pftprc"
#define PFTPSYSTEMRC         "/etc/pftp.conf"
/* link script definitions */
#define PFTPMAKELINKS        "PFTP_MAKE_LINKS"
#define PFTPLINKSCRIPTINIT   "#!/bin/sh\n"
#define PFTPSCRIPTINITLEN    10
#define PFTPLINKPROG         "ln -s %s %s\n"

#if defined HAVE_GLIB || defined ULTRIX || defined NEXTSTEP || defined OSF1
#include <errno.h>
#include <pwd.h>
#include <sys/types.h>
#endif

struct status_type {
#ifndef USE_POSIX_THREAD
   unsigned isparent: 1;
#endif
   unsigned _STANDARD_INPUT_: 1; /* standard input */
   unsigned FORCE_STDOUT: 1; /* send received files to stdout */
   unsigned _RECURS_: 1; /* send directories recursivly */
   unsigned OVERWRITE: 2; /* overwrite all existing files */
   unsigned _PFTP_DAEMON_: 2; /* run as daemon: 1) keep running 2) exit after first connect */
   unsigned lol: 1; /* output '\r' */
   unsigned lognames: 1; /* log names when running as inetd server/ daemon */
   unsigned delete_info: 1; /* delete info file */
   unsigned READFROMSTDIN: 1; /* read file names from stdin */
   unsigned QT_FRONTEND: 1; /* client is running by the qt front end `qpftp' */
   unsigned ttys: 1; /* 0 and 1 are ttys */
   unsigned exit_afc: 1; /* exit after first connection */
   unsigned verbose: 1; /* output some more information */
   unsigned single_connection: 1; /* accept only one connection at a time */
   unsigned use_udp: 1; /* using use datagram protocol (UDP) */
   unsigned first: 2;
   unsigned show_bytes: 1; /* display read in bytes during UDP transmission */
#ifdef Linux
   unsigned new_line: 1;
#endif
#ifdef USE_POSIX_THREAD
   unsigned free: 1;
   unsigned RECEIVING: 1;
#endif
   /* Multicast members */
#if defined IPV6_MULTICAST_LOOP || defined IP_MULTICAST_LOOP
   unsigned multicast: 1; /* pftp sends resp. receives multicasted data */
   unsigned int mloop; /* IPV6_MULTICAST_LOOP resp. IP_MULTICAST_LOOP */
   int mttl; /* hop limit resp. time to live */
   char *_INTERFACE_;
   char *_MULTICASTGROUP_;
#endif
   /* End of multicast members */
   short _SKIP_;    /* skip all existing files */
   short _OPTIONS_; /* number of options */
   size_t _STDIN_BUFSIZ_; /* size of stdin buffer */
   size_t bsize; /* size of net buffer */
   char *_HOSTNAME_;
   char *home; /* home directory */
   char *version; /* version of server */
   char *sstatus; /* status of server */
   char *SENDMAKELINKS; /* don't follow symbolic links; send script file */
   int ns; /* client/server port to read from */
   int ws; /* client/server port to write to */
   double mul;
   unsigned long throw_away;
   double testbsize;
   unsigned long _BANDWIDTH_; /* sets the bandwidth in bytes */
   FILE *fp;
   FILE *rp;
   int pid;
   char *rlogin;
   char *rpass;
   char *from;
   char *pw_name; 
   char *pw_passwd; 
#ifdef HAVE_INET6
   int priority;
   struct addrinfo *adin;
   struct sockaddr_in6 sin;
#else
   struct sockaddr_in sin;
#endif
#ifdef USE_POSIX_THREAD
   int output;
#endif
   double start;
   char *pw_dir; 
   uid_t pw_uid;
   gid_t pw_gid;
   int sinlen;
   unsigned int maxclients;
   unsigned int no_clients;
   unsigned int usefilter; /* flag for option '-f' */
   char REMOTEHOSTNAME[256]; /* name of current remote client host */
};

struct client_type {
   unsigned is_subdir: 1; /* determine if sending() was called recursively */
   char **argv;
   int argc;
};

#define MEM_CHECK(ALLOC_MEM) \
        if (!ALLOC_MEM) { \
           fprintf(stderr, "** calloc: Cannot allocate enough memory.\n"); \
           exit(1); \
        }
#endif /* STATUS_STRUCTURE_H */
